From stdpp Require Export gmultiset strings.
From iris.algebra Require Import frac.
From iris.base_logic.lib Require Export invariants.
From iris.bi Require Import fractional.
From orc11 Require Import base.
From gpfsl.base_logic Require Export vprop.

From iris.proofmode Require Import proofmode.
From iris.prelude Require Import options.

Definition lftN : namespace := nroot .@ "lft".

Implicit Types (Lat : latticeT) (E : coPset).

Module Type lifetime_sig.
  (** CMRAs needed by the lifetime logic  *)
  (* We can't instantie the module parameters with inductive types, so we
     have aliases here. *)
  (** userE is a (disjoint) mask that is available in the "consequence" view
      shift of borrow accessors. *)
  Parameter lftG' : ∀ (Σ : gFunctors) (Lat : latticeT) (userE : coPset), Set.
  Global Notation lftG := lftG'.
  Existing Class lftG'.
  Parameter lftPreG' : ∀ (Σ : gFunctors) (Lat : latticeT), Set.
  Global Notation lftPreG := lftPreG'.
  Existing Class lftPreG'.

  (** * Definitions *)
  Parameter lft : Type.
  Parameter static : lft.
  Global Declare Instance lft_intersect : Meet lft.

  Section defs.
  Notation monPred Σ Lat := (monPredI (lat_bi_index Lat) (iPropI Σ)).

  Parameter lft_ctx : ∀ `{!invGS Σ, !lftG Σ Lat userE}, iProp Σ.

  Parameter lft_tok : ∀ `{!lftG Σ Lat userE} (q : Qp) (κ : lft), monPred Σ Lat.
  Parameter lft_dead : ∀ `{!lftG Σ Lat userE} (κ : lft), monPred Σ Lat.

  Parameter lft_incl :
    ∀ `{!invGS Σ, !lftG Σ Lat userE} (κ κ' : lft), monPred Σ Lat.
  Parameter bor :
    ∀ `{!invGS Σ, !lftG Σ Lat userE} (κ : lft) (P : monPred Σ Lat), monPred Σ Lat.

  Parameter bor_idx : Type.
  Parameter idx_bor_own : ∀ `{!lftG Σ Lat userE} (i : bor_idx), monPred Σ Lat.
  Parameter idx_bor :
    ∀ `{!invGS Σ, !lftG Σ Lat userE} (κ : lft) (i : bor_idx) (P : monPred Σ Lat),
      monPred Σ Lat.

  Parameter in_at_bor :
    ∀ `{!invGS Σ, !lftG Σ Lat userE} (κ : lft) (P : monPred Σ Lat), monPred Σ Lat.
  End defs.

  (** Our lifetime creation lemma offers allocating a lifetime that is defined
  by a [positive] in some given infinite set. This operation converts the
  [positive] to a lifetime. *)
  Parameter positive_to_lft : positive → lft.

  (** * Notation *)
  Notation "q .[ κ ]" := (lft_tok q κ)
      (format "q .[ κ ]", at level 2, left associativity) : bi_scope.
  Notation "[† κ ]" := (lft_dead κ) (format "[† κ ]"): bi_scope.

  Notation "&{ κ }" := (bor κ) (format "&{ κ }") : bi_scope.
  Notation "&{ κ , i }" := (idx_bor κ i) (format "&{ κ , i }") : bi_scope.
  Notation "&in_at{ κ }" := (in_at_bor κ) (format "&in_at{ κ }") : bi_scope.

  Infix "⊑" := lft_incl (at level 70) : bi_scope.

  Section properties.
  Context `{!invGS Σ, !lftG Σ Lat userE}.

  (** * Instances *)
  Global Declare Instance lft_inhabited : Inhabited lft.
  Global Declare Instance bor_idx_inhabited : Inhabited bor_idx.

  Global Declare Instance lft_intersect_comm : Comm (A:=lft) eq (⊓).
  Global Declare Instance lft_intersect_assoc : Assoc (A:=lft) eq (⊓).
  Global Declare Instance lft_intersect_inj_1 (κ : lft) : Inj eq eq (κ ⊓.).
  Global Declare Instance lft_intersect_inj_2 (κ : lft) : Inj eq eq (.⊓ κ).
  Global Declare Instance lft_intersect_left_id : LeftId eq static meet.
  Global Declare Instance lft_intersect_right_id : RightId eq static meet.

  Global Declare Instance lft_ctx_persistent : Persistent lft_ctx.
  Global Declare Instance lft_dead_persistent κ : Persistent [†κ].
  Global Declare Instance lft_incl_persistent κ κ' : Persistent (κ ⊑ κ').
  Global Declare Instance idx_bor_persistent κ i P : Persistent (&{κ, i}P).
  Global Declare Instance in_at_bor_persistent κ P : Persistent (&in_at{κ}P).

  Global Declare Instance lft_tok_timeless q κ : Timeless q.[κ].
  Global Declare Instance lft_dead_timeless κ : Timeless [†κ].
  Global Declare Instance idx_bor_own_timeless i : Timeless (idx_bor_own i).

  Global Declare Instance idx_bor_params : Params (@idx_bor) 6.
  Global Declare Instance bor_params : Params (@bor) 5.
  Global Declare Instance in_at_bor_params : Params (@in_at_bor) 5.

  Global Declare Instance bor_ne κ n : Proper (dist n ==> dist n) (bor κ).
  Global Declare Instance bor_contractive κ : Contractive (bor κ).
  Global Declare Instance bor_proper κ : Proper ((⊣⊢) ==> (⊣⊢)) (bor κ).
  Global Declare Instance idx_bor_ne κ i n : Proper (dist n ==> dist n) (idx_bor κ i).
  Global Declare Instance idx_bor_contractive κ i : Contractive (idx_bor κ i).
  Global Declare Instance idx_bor_proper κ i : Proper ((≡) ==> (≡)) (idx_bor κ i).
  Global Declare Instance in_at_bor_ne κ n : Proper (dist n ==> dist n) (in_at_bor κ).
  Global Declare Instance in_at_bor_contractive κ : Contractive (in_at_bor κ).
  Global Declare Instance in_at_bor_proper κ : Proper ((≡) ==> (≡)) (in_at_bor κ).

  Global Declare Instance lft_tok_fractional κ : Fractional (λ q, q.[κ])%I.
  Global Declare Instance lft_tok_as_fractional κ q :
    AsFractional q.[κ] (λ q, q.[κ])%I q.
  Global Declare Instance frame_lft_tok p κ q1 q2 RES :
    FrameFractionalHyps p q1.[κ] (λ q, q.[κ])%I RES q1 q2 →
    Frame p q1.[κ] q2.[κ] RES | 5.

  Global Declare Instance positive_to_lft_inj : Inj eq eq positive_to_lft.

  (** * Laws *)
  Parameter lft_tok_sep : ∀ q κ1 κ2, q.[κ1] ∗ q.[κ2] ⊣⊢ q.[κ1 ⊓ κ2].
  Parameter lft_tok_split_obj :
    ∀ `{LatBottom Lat} q1 q2 κ, (q1 + q2).[κ] -∗ q1.[κ] ∗ <obj> q2.[κ].
  Parameter lft_dead_or : ∀ κ1 κ2, [†κ1] ∨ [†κ2] ⊣⊢ [† κ1 ⊓ κ2].
  Parameter lft_tok_dead : ∀ q κ, q.[κ] -∗ [† κ] -∗ False.
  Parameter lft_tok_static : ∀ q, ⊢ q.[static].
  Parameter lft_dead_static : [† static] -∗ False.
  Parameter lft_intersect_static_cancel_l : ∀ κ κ', κ ⊓ κ' = static → κ = static.

  (** Create a lifetime in some given set of names [P]. This lemma statement
     requires exposing [atomic_lft], because [P] restricted to the image of
     [atomic_to_lft] might well not be infinite. *)
  Parameter lft_create_strong : ∀ `{LatBottom Lat} P E, pred_infinite P → ↑lftN ⊆ E →
    ⎡lft_ctx⎤ ={E}=∗
    ∃ p : positive, let κ := positive_to_lft p in ⌜P p⌝ ∗
         (1).[κ] ∗ □ ((1).[κ] ={↑lftN ∪ userE}[userE]▷=∗ [†κ]).
  Parameter bor_create : ∀ `{LatBottom Lat} E κ P,
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ ▷ P ={E}=∗ &{κ} P ∗ ([†κ] ={E}=∗ ▷ P).
  Parameter bor_fake : ∀ `{LatBottom Lat} E κ P,
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ <subj>[†κ] ={E}=∗ &{κ}P.

  Parameter bor_sep : ∀ `{LatBottom Lat} E κ P Q, ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗
    &{κ}(P ∗ Q) ={E}=∗ &{κ}P ∗ &{κ}Q.
  Parameter bor_combine : ∀ `{LatBottom Lat} E κ P Q,
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ}P -∗ &{κ}Q ={E}=∗  &{κ}(P ∗ Q).

  Parameter bor_unfold_idx : ∀ κ P, (&{κ}P)%I ⊣⊢ ∃ i, (&{κ,i}P)%I ∗ (idx_bor_own i).

  Parameter idx_bor_shorten : ∀ κ κ' i P, κ ⊑ κ' -∗ &{κ',i}P -∗ &{κ,i}P.
  Parameter idx_bor_iff : ∀ κ i P P', ▷ □ (P ↔ P') -∗ &{κ,i}P -∗ &{κ,i}P'.

  Parameter idx_bor_unnest : ∀ `{LatBottom Lat} E κ κ' i P,
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ,i}P -∗ &{κ'}(idx_bor_own i) ={E}=∗ &{κ ⊓ κ'}P.
  Parameter in_at_bor_shorten : ∀ κ κ' P, κ ⊑ κ' -∗ &in_at{κ'}P -∗ &in_at{κ}P.

  Parameter in_at_bor_iff : ∀ κ P P',
    ▷ □ (P ↔ P') -∗ &in_at{κ}P -∗ &in_at{κ}P'.
  Parameter in_at_bor_share : ∀ E κ P, &{κ}P ={E}=∗ &in_at{κ}P.

  Parameter idx_bor_acc : ∀ E q κ i P Vtok Vbor, ↑lftN ⊆ E →
    lft_ctx -∗ &{κ,i}P Vtok -∗ idx_bor_own i Vbor -∗ q.[κ] Vtok ={E}=∗
      ∃ V, ⌜V ⊑ Vbor ⊔ Vtok⌝ ∗ ▷ P V ∗
        (∀ Vtok', ⌜Vtok ⊑ Vtok'⌝ -∗ ▷ P (Vtok' ⊔ V)
                  ={E}=∗ idx_bor_own i (Vtok' ⊔ V) ∗ q.[κ] Vtok').
  (* We have to expose κ' here as without [lftN] in the mask of the Q-to-P view
     shift, we cannot turn [†κ'] into [†κ].

     The reason why [†κ'] is under a subjective modality is because of what
     happens when a non-atomic lifetime dies (a non-atomic lifetime is the
     intersection of other lifetimes). Indeed, a non-atomic lifetime can die
     "several times", once for each component, and each of these death can
     occur at non-comparable views. The only relevant view that we would pass
     to the consequence view-shift is the view used for the inheritance of
     the borrow, but that view is not necessarily well-defined since the
     borrow can be the result of a merging of other borrows.

     TODO : It would be great to get rid of that modality, because this should
     allow giving a model to GhostCell under weak memory. See the comment in
     ghostcell.v in the master SC branch.
 *)
  Parameter bor_acc_strong : ∀ E q κ P, ↑lftN ⊆ E →
    ⎡lft_ctx⎤ -∗ &{κ}P -∗ q.[κ] ={E}=∗ ∃ κ', (κ ⊑ κ') ∗ ▷ P ∗
      ∀ Q, ▷ (▷ Q -∗ <subj>[†κ'] ={userE}=∗ ▷ P) -∗ ▷ Q ={E}=∗ &{κ'}Q ∗ q.[κ].
  (* Atomic accessor for full borrows.

     In the context of view predicates, this accessor is a bit
     different when compared to the sequentially consistent case.
     Indeed, first, even if we can change the underlying predicate
     (i.e., going from P to Q), we must preserve the underlying
     resource, so that we need to preserve an existentially quantified
     assertion P'.

     Note that the "consequence" view shift is only executed if the
     borrow is actually opened afterwards (if the borrow is not
     opened, then P' still holds and can be used in the
     inheritance). Hence, we actually get that the consequence view
     shift happens at a view which is larger than the current view.
     See the comment in accessors.v for why the consequence VS only
     gets the dead token under the subjectively modality. *)
  Parameter bor_acc_atomic_strong : ∀ E κ P, ↑lftN ⊆ E →
    ⎡lft_ctx⎤ -∗ &{κ}P ={E,E∖↑lftN}=∗
      (∃ P' κ', κ ⊑ κ' ∗ ▷ (⎡P'⎤ ∧ P) ∗
         ∀ Q,
           ▷ (▷ Q -∗ <subj>[†κ'] ={userE}=∗ ▷ P) -∗
           ▷ (⎡P'⎤ ∧ Q)
           ={E∖↑lftN,E}=∗ &{κ'}Q) ∨
      (∃ κ', (κ ⊑ κ') ∗ <subj>[†κ'] ∗ |={E∖↑lftN,E}=> True).
  Parameter in_at_bor_acc : ∀ E q κ P, ↑lftN ⊆ E →
    ⎡lft_ctx⎤ -∗ &in_at{κ}P -∗ q.[κ] ={E,E∖↑lftN}=∗ ∃ Vb,
      (monPred_in Vb → ▷ P) ∗ ((monPred_in Vb → ▷ P) ={E∖↑lftN,E}=∗ q.[κ]).

  (* Because Coq's module system is horrible, we have to repeat properties of lft_incl here
     unless we want to prove them twice (both here and in model.primitive) *)
  Parameter lft_intersect_acc : ∀ κ κ' q q', q.[κ] -∗ q'.[κ'] -∗
    ∃ q'', q''.[κ ⊓ κ'] ∗ (q''.[κ ⊓ κ'] -∗ q.[κ] ∗ q'.[κ']).
  Parameter lft_intersect_incl_l : ∀ κ κ', ⊢ κ ⊓ κ' ⊑ κ.
  Parameter lft_intersect_incl_r : ∀ κ κ', ⊢ κ ⊓ κ' ⊑ κ'.
  Parameter lft_incl_refl : ∀ κ, ⊢ κ ⊑ κ.
  Parameter lft_incl_trans : ∀ κ κ' κ'', κ ⊑ κ' -∗ κ' ⊑ κ'' -∗ κ ⊑ κ''.
  Parameter lft_incl_glb : ∀ κ κ' κ'', κ ⊑ κ' -∗ κ ⊑ κ'' -∗ κ ⊑ κ' ⊓ κ''.
  Parameter lft_intersect_mono : ∀ κ1 κ1' κ2 κ2',
    κ1 ⊑ κ1' -∗ κ2 ⊑ κ2' -∗ κ1 ⊓ κ2 ⊑ κ1' ⊓ κ2'.
  Parameter lft_incl_acc : ∀ E κ κ' q,
    ↑lftN ⊆ E → κ ⊑ κ' -∗ q.[κ] ={E}=∗ ∃ q', q'.[κ'] ∗ (q'.[κ'] ={E}=∗ q.[κ]).
  Parameter lft_incl_intro : ∀ κ κ',
    □ (∀ q, q.[κ] ={↑lftN}=∗ ∃ q', q'.[κ'] ∗ (q'.[κ'] ={↑lftN}=∗ q.[κ]))
      -∗ κ ⊑ κ'.
  End properties.

  Parameter lftΣ : latticeT → gFunctors.
  Global Declare Instance subG_lftPreG Lat Σ : subG (lftΣ Lat) Σ → lftPreG Σ Lat.

  Parameter lft_init : ∀ `{!invGS Σ, !lftPreG Σ Lat} E userE,
      ↑lftN ⊆ E → ↑lftN ## userE →
      ⊢ |={E}=> ∃ _ : lftG Σ Lat userE, lft_ctx.
End lifetime_sig.
